/*

Copyright (c) 2012 Red Hat Incorporated.
All rights reserved.

Redistribution and use in source and binary forms, with or without 
modification, are permitted provided that the following conditions are met: 

    Redistributions of source code must retain the above copyright 
    notice, this list of conditions and the following disclaimer.

    Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.

    The name of Red Hat Incorporated may not be used to endorse 
    or promote products derived from this software without specific 
    prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED.  IN NO EVENT SHALL RED HAT INCORPORATED BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND 
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 

*/

#include "vregs.h"
	
	.text

	.global	_memcmp
	.type	_memcmp, @function
_memcmp:
	;; [sp+4].HI = dest
	;; [sp+6].HI = src
	;; [sp+8].HI = count

	movw	ax, [sp+4]
	movw	de, ax
	movw	ax, [sp+6]
	movw	hl, ax
	mov	a, [sp+8]
	or	a, l
	or	a, e
	mov1	cy, a.0
	bnc	$Lmemcmp_words

Lmemcmp_bytes:
	movw	ax, [sp+8]
	movw	bc, ax		; so we can INC and use BZ
	movw	ax, #0xffff
	subw	ax, bc
	movw	bc, ax

Ltop:	
	.macro	one_iter,N

	inc	c
	bz	$Lincb_b
Lincb_ret_\N:

	mov	a, [de]
	cmp	a, [hl]
	bnz	$Lmismatch

	incw	de
	incw	hl
	.endm

	one_iter	0
	one_iter	1
	one_iter	2
	one_iter	3

	br	$Ltop

Lincb_b:
	inc	b
	bnz	$Lincb_ret_0
	br	$Lend


Lend:
	movw	r8, #0
	ret

Lmismatch:
	sub	a, [hl]
	mov	r8, a
	sar	a, 7
	mov	r9, a
	ret

Lmemcmp_words:
	movw	ax, [sp+8]
	shrw	ax, 1
	movw	bc, ax		; so we can INC and use BZ
	movw	ax, #0xffff
	subw	ax, bc
	movw	bc, ax

Ltopw:	
	.macro	one_iterw,N

	inc	c
	bz	$Lincr_b_\N
Lincr_ret_\N:

	movw	ax, [de+\N]
	cmpw	ax, [hl+\N]
	bnz	$Lmismatch_words_\N
	.endm

	one_iterw	0
	one_iterw	2
	one_iterw	4
	one_iterw	6

	movw	ax, hl
	addw	ax, #8
	movw	hl, ax
	movw	ax, de
	addw	ax, #8
	movw	de, ax

	br	$Ltopw

	.macro	incr_b,N
Lincr_b_\N:	
	inc	b
	bz	$Lend
	br	$Lincr_ret_\N
	.endm
	incr_b	0
	incr_b	2
	incr_b	4
	incr_b	6
	
Lmismatch_words_6:
	movw	ax, de
	addw	ax, #6
	movw	de, ax
	movw	ax, hl
	addw	ax, #6
	movw	hl, ax
	movw	ax, [de]
	br	$Lmismatch_words_0

Lmismatch_words_4:
	movw	ax, de
	addw	ax, #4
	movw	de, ax
	movw	ax, hl
	addw	ax, #4
	movw	hl, ax
	movw	ax, [de]
	br	$Lmismatch_words_0

Lmismatch_words_2:
	incw	hl
	incw	hl
	incw	de
	incw	de
	
Lmismatch_words_0:
	cmps	x, [hl+0]
	bz	$Lmismatch_words_even
	bc	$Lmismatch_words_cy
	br	$Lmismatch_words_nc

Lmismatch_words_even:
	cmp	a, [hl+1]
	bnc	$Lmismatch_words_nc

Lmismatch_words_cy:
	movw	r8, #0xffff
	ret
Lmismatch_words_nc:
	movw	r8, #1
	ret
	
	.size	_memcmp, .-_memcmp

