/*

Copyright (c) 2012 Red Hat Incorporated.
All rights reserved.

Redistribution and use in source and binary forms, with or without 
modification, are permitted provided that the following conditions are met: 

    Redistributions of source code must retain the above copyright 
    notice, this list of conditions and the following disclaimer.

    Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.

    The name of Red Hat Incorporated may not be used to endorse 
    or promote products derived from this software without specific 
    prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED.  IN NO EVENT SHALL RED HAT INCORPORATED BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND 
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 

*/

#include "vregs.h"
	
.text
	.global	_memcpy
	.type	_memcpy, @function
_memcpy:
	;; [sp+4].HI = dest
	;; [sp+6].HI = src
	;; [sp+8].HI = count

	movw	ax, [sp+8]
	movw	r8, ax
	cmpw	ax, #0
	sknz
	br	$!done

	movw	ax, [sp+4]
	movw	hl, ax
	movw	ax, [sp+6]
	movw	de, ax

	;; mismatched alignment - use byte moves
	mov	a, e
	xor	a, l
	mov1	cy,a.0
	bc	$top_byte
	
	;; all odd, one byte move aligns them
	mov	a, e
	and	a, l
	mov1	cy,a.0
	bnc	$both_aligned

	;; both are unaligned, align them
	mov	a, [de]
	mov	[hl], a
	incw	hl
	incw	de
	decw	r8

both_aligned:	
	movw	ax, r8
	cmpw	ax, #8
	bnh	$word_loop
	movw	ax, r8
	shrw	ax, 3
	movw	r10, ax
	mov	a, r8
	and	a, #7
	mov	r8, a
	mov	r9, #0

top_block:
	movw	ax, [de]
	movw	[hl],ax
	movw	ax, [de+2]
	movw	[hl+2],ax
	movw	ax, [de+4]
	movw	[hl+4],ax
	movw	ax, [de+6]
	movw	[hl+6],ax

	movw	ax, hl
	addw	ax, #8
	movw	hl, ax
	movw	ax, de
	addw	ax, #8
	movw	de, ax
	movw	ax, r10
	subw	ax, #1
	movw	r10, ax
	bnz	$top_block

word_loop:	
	movw	ax, r8
	cmpw	ax, #2
	bnh	$simple_byte
	movw	ax, r8
	shrw	ax, 1
	movw	r10, ax

top_word:
	movw	ax, [de]
	movw	[hl],ax
	incw	hl
	incw	hl
	incw	de
	incw	de
	movw	ax, r10
	subw	ax, #1
	movw	r10, ax
	bnz	$top_word

	mov1	cy, r8.0
	bnc	$done
	mov	a, [de]
	mov	[hl], a
	br	$done

	;; one of the two addresses is unaligned.
top_byte:
	movw	ax, r8
	shrw	ax, 1
	shrw	ax, 1
	cmpw	ax, #0
	bz	$simple_byte
	movw	r10, ax
	mov	a, r8
	and	a, #3
	mov	r8, a
	mov	r9, #0

fast_byte:
	mov	a, [de]
	mov	[hl],a
	incw	hl
	incw	de
	mov	a, [de]
	mov	[hl],a
	incw	hl
	incw	de
	mov	a, [de]
	mov	[hl],a
	incw	hl
	incw	de
	mov	a, [de]
	mov	[hl],a
	incw	hl
	incw	de
	movw	ax, r10
	subw	ax, #1
	movw	r10, ax
	bnz	$fast_byte

simple_byte:
	movw	ax, r8
	cmpw	ax, #0
	bz	$done
simple_byte_1:	
	mov	a, [de]
	mov	[hl],a
	incw	hl
	incw	de
	movw	ax, r8
	subw	ax, #1
	movw	r8, ax
	bnz	$simple_byte_1
	
done:
	movw	ax, [sp+4]
	movw	r8, ax
	ret

	.size	_memcpy, .-_memcpy
